@@ -17,9 +17,18 @@
 
 from __future__ import print_function
 
+from datetime import datetime
+
+import json
+import shutil
+
+from azurelinuxagent.common.event import *
 from azurelinuxagent.common.protocol.hostplugin import *
+from azurelinuxagent.common.protocol.metadata import *
 from azurelinuxagent.common.protocol.wire import *
+from azurelinuxagent.common.utils.fileutil import *
 from azurelinuxagent.ga.update import *
+
 from tests.tools import *
 
 NO_ERROR = {
@@ -28,6 +37,18 @@ NO_ERROR = {
     "was_fatal" : False
 }
 
+FATAL_ERROR = {
+    "last_failure" : 42.42,
+    "failure_count" : 2,
+    "was_fatal" : True
+}
+
+SENTINEL_ERROR = {
+    "last_failure" : 0.0,
+    "failure_count" : 0,
+    "was_fatal" : True
+}
+
 WITH_ERROR = {
     "last_failure" : 42.42,
     "failure_count" : 2,
@@ -128,7 +149,9 @@ class UpdateTestCase(AgentTestCase):
 
     def create_error(self, error_data=NO_ERROR):
         with self.get_error_file(error_data) as path:
-            return GuestAgentError(path.name)
+            err = GuestAgentError(path.name)
+            err.load()
+            return err
 
     def copy_agents(self, *agents):
         if len(agents) <= 0:
@@ -137,19 +160,25 @@ class UpdateTestCase(AgentTestCase):
             fileutil.copy_file(agent, to_dir=self.tmp_dir)
         return
 
-    def expand_agents(self):
+    def expand_agents(self, mark_optional=False):
         for agent in self.agent_pkgs():
-            zipfile.ZipFile(agent).extractall(os.path.join(
-                self.tmp_dir,
-                fileutil.trim_ext(agent, "zip")))
+            path = os.path.join(self.tmp_dir, fileutil.trim_ext(agent, "zip"))
+            zipfile.ZipFile(agent).extractall(path)
+            if mark_optional:
+                src = os.path.join(data_dir, 'ga', 'supported.json')
+                dst = os.path.join(path, 'supported.json')
+                shutil.copy(src, dst)
+
+                dst = os.path.join(path, 'error.json')
+                fileutil.write_file(dst, json.dumps(SENTINEL_ERROR))
         return
 
-    def prepare_agent(self, version):
+    def prepare_agent(self, version, mark_optional=False):
         """
         Create a download for the current agent version, copied from test data
         """
         self.copy_agents(get_agent_pkgs()[0])
-        self.expand_agents()
+        self.expand_agents(mark_optional=mark_optional)
 
         versions = self.agent_versions()
         src_v = FlexibleVersion(str(versions[0]))
@@ -214,6 +243,64 @@ class UpdateTestCase(AgentTestCase):
         return dst_v
 
 
+class TestSupportedDistribution(UpdateTestCase):
+    def setUp(self):
+        UpdateTestCase.setUp(self)
+        self.sd = SupportedDistribution({
+                    'slice':10,
+                    'versions': ['^Ubuntu,16.10,yakkety$']})
+
+    def test_creation(self):
+        self.assertRaises(TypeError, SupportedDistribution)
+        self.assertRaises(UpdateError, SupportedDistribution, None)
+
+        self.assertEqual(self.sd.slice, 10)
+        self.assertEqual(self.sd.versions, ['^Ubuntu,16.10,yakkety$'])
+
+    @patch('platform.linux_distribution')
+    def test_is_supported(self, mock_dist):
+        mock_dist.return_value = ['Ubuntu', '16.10', 'yakkety']
+        self.assertTrue(self.sd.is_supported)
+
+        mock_dist.return_value = ['something', 'else', 'entirely']
+        self.assertFalse(self.sd.is_supported)
+
+    @patch('azurelinuxagent.ga.update.datetime')
+    def test_in_slice(self, mock_dt):
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 5))
+        self.assertTrue(self.sd.in_slice)
+
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 42))
+        self.assertFalse(self.sd.in_slice)
+
+
+class TestSupported(UpdateTestCase):
+    def setUp(self):
+        UpdateTestCase.setUp(self)
+        self.sp = Supported(os.path.join(data_dir, 'ga', 'supported.json'))
+        self.sp.load()
+
+    def test_creation(self):
+        self.assertRaises(TypeError, Supported)
+        self.assertRaises(UpdateError, Supported, None)
+
+    @patch('platform.linux_distribution')
+    def test_is_supported(self, mock_dist):
+        mock_dist.return_value = ['Ubuntu', '16.10', 'yakkety']
+        self.assertTrue(self.sp.is_supported)
+
+        mock_dist.return_value = ['something', 'else', 'entirely']
+        self.assertFalse(self.sp.is_supported)
+
+    @patch('platform.linux_distribution', return_value=['Ubuntu', '16.10', 'yakkety'])
+    @patch('azurelinuxagent.ga.update.datetime')
+    def test_in_slice(self, mock_dt, mock_dist):
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 5))
+        self.assertTrue(self.sp.in_slice)
+
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 42))
+        self.assertFalse(self.sp.in_slice)
+
 class TestGuestAgentError(UpdateTestCase):
     def test_creation(self):
         self.assertRaises(TypeError, GuestAgentError)
@@ -221,6 +308,7 @@ class TestGuestAgentError(UpdateTestCase):
 
         with self.get_error_file(error_data=WITH_ERROR) as path:
             err = GuestAgentError(path.name)
+            err.load()
             self.assertEqual(path.name, err.path)
         self.assertNotEqual(None, err)
 
@@ -232,6 +320,7 @@ class TestGuestAgentError(UpdateTestCase):
     def test_clear(self):
         with self.get_error_file(error_data=WITH_ERROR) as path:
             err = GuestAgentError(path.name)
+            err.load()
             self.assertEqual(path.name, err.path)
         self.assertNotEqual(None, err)
 
@@ -241,18 +330,18 @@ class TestGuestAgentError(UpdateTestCase):
         self.assertEqual(NO_ERROR["was_fatal"], err.was_fatal)
         return
 
-    def test_load_preserves_error_state(self):
-        with self.get_error_file(error_data=WITH_ERROR) as path:
+    def test_is_sentinel(self):
+        with self.get_error_file(error_data=SENTINEL_ERROR) as path:
             err = GuestAgentError(path.name)
-            self.assertEqual(path.name, err.path)
-        self.assertNotEqual(None, err)
-
-        with self.get_error_file(error_data=NO_ERROR):
             err.load()
-        self.assertEqual(WITH_ERROR["last_failure"], err.last_failure)
-        self.assertEqual(WITH_ERROR["failure_count"], err.failure_count)
-        self.assertEqual(WITH_ERROR["was_fatal"], err.was_fatal)
-        return
+            self.assertTrue(err.is_blacklisted)
+            self.assertTrue(err.is_sentinel)
+        
+        with self.get_error_file(error_data=FATAL_ERROR) as path:
+            err = GuestAgentError(path.name)
+            err.load()
+            self.assertTrue(err.is_blacklisted)
+            self.assertFalse(err.is_sentinel)
 
     def test_save(self):
         err1 = self.create_error()
@@ -274,15 +363,6 @@ class TestGuestAgentError(UpdateTestCase):
         # Agent failed >= MAX_FAILURE, it should be blacklisted
         self.assertTrue(err.is_blacklisted)
         self.assertEqual(MAX_FAILURE, err.failure_count)
-
-        # Clear old failure does not clear recent failure
-        err.clear_old_failure()
-        self.assertTrue(err.is_blacklisted)
-
-        # Clear does remove old, outdated failures
-        err.last_failure -= RETAIN_INTERVAL * 2
-        err.clear_old_failure()
-        self.assertFalse(err.is_blacklisted)
         return
 
     def test_mark_failure_permanent(self):
@@ -320,20 +400,21 @@ class TestGuestAgent(UpdateTestCase):
         self.agent_path = os.path.join(self.tmp_dir, get_agent_name())
         return
 
-    def tearDown(self):
-        self.remove_agents()
-        return
-
     def test_creation(self):
         self.assertRaises(UpdateError, GuestAgent, "A very bad file name")
         n = "{0}-a.bad.version".format(AGENT_NAME)
         self.assertRaises(UpdateError, GuestAgent, n)
 
+        self.expand_agents()
+
         agent = GuestAgent(path=self.agent_path)
         self.assertNotEqual(None, agent)
         self.assertEqual(get_agent_name(), agent.name)
         self.assertEqual(get_agent_version(), agent.version)
 
+        self.assertFalse(agent._is_optional)
+        self.assertFalse(agent._in_slice)
+
         self.assertEqual(self.agent_path, agent.get_agent_dir())
 
         path = os.path.join(self.agent_path, AGENT_MANIFEST_FILE)
@@ -347,13 +428,14 @@ class TestGuestAgent(UpdateTestCase):
         self.assertEqual(path, agent.get_agent_pkg_path())
 
         self.assertTrue(agent.is_downloaded)
-        # Note: Agent will get blacklisted since the package for this test is invalid
-        self.assertTrue(agent.is_blacklisted)
-        self.assertFalse(agent.is_available)
+        self.assertFalse(agent.is_blacklisted)
+        self.assertTrue(agent.is_available)
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_clear_error(self, mock_ensure):
+    def test_clear_error(self, mock_downloaded):
+        self.expand_agents()
+
         agent = GuestAgent(path=self.agent_path)
         agent.mark_failure(is_fatal=True)
 
@@ -370,7 +452,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_is_available(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_is_available(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
 
         self.assertFalse(agent.is_available)
@@ -382,7 +465,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_is_blacklisted(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_is_blacklisted(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(agent.is_blacklisted)
 
@@ -396,16 +480,56 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_is_downloaded(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_is_downloaded(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(agent.is_downloaded)
         agent._unpack()
         self.assertTrue(agent.is_downloaded)
         return
 
-    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_mark_failure(self, mock_ensure):
+    @patch('platform.linux_distribution', return_value=['Ubuntu', '16.10', 'yakkety'])
+    @patch('azurelinuxagent.ga.update.GuestAgent._enable')
+    def test_is_optional(self, mock_enable, mock_dist):
+        self.expand_agents(mark_optional=True)
         agent = GuestAgent(path=self.agent_path)
+
+        self.assertTrue(agent.is_blacklisted)
+        self.assertTrue(agent._is_optional)
+
+    @patch('platform.linux_distribution', return_value=['Ubuntu', '16.10', 'yakkety'])
+    @patch('azurelinuxagent.ga.update.datetime')
+    def test_in_slice(self, mock_dt, mock_dist):
+        self.expand_agents(mark_optional=True)
+        agent = GuestAgent(path=self.agent_path)
+
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 5))
+        self.assertTrue(agent._in_slice)
+
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 42))
+        self.assertFalse(agent._in_slice)
+
+    @patch('platform.linux_distribution', return_value=['Ubuntu', '16.10', 'yakkety'])
+    @patch('azurelinuxagent.ga.update.datetime')
+    def test_enable(self, mock_dt, mock_dist):
+        mock_dt.utcnow = Mock(return_value=datetime(2017, 1, 1, 0, 0, 5))
+
+        self.expand_agents(mark_optional=True)
+        agent = GuestAgent(path=self.agent_path)
+
+        self.assertFalse(agent.is_blacklisted)
+        self.assertFalse(agent._is_optional)
+
+        # Ensure the new state is preserved to disk
+        agent = GuestAgent(path=self.agent_path)
+        self.assertFalse(agent.is_blacklisted)
+        self.assertFalse(agent._is_optional)
+
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_mark_failure(self, mock_loaded, mock_downloaded):
+        agent = GuestAgent(path=self.agent_path)
+
         agent.mark_failure()
         self.assertEqual(1, agent.error.failure_count)
 
@@ -415,7 +539,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_unpack(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_unpack(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(os.path.isdir(agent.get_agent_dir()))
         agent._unpack()
@@ -424,7 +549,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_unpack_fail(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_unpack_fail(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(os.path.isdir(agent.get_agent_dir()))
         os.remove(agent.get_agent_pkg_path())
@@ -432,7 +558,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_load_manifest(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_load_manifest(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         agent._unpack()
         agent._load_manifest()
@@ -441,7 +568,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_load_manifest_missing(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_load_manifest_missing(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(os.path.isdir(agent.get_agent_dir()))
         agent._unpack()
@@ -450,7 +578,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_load_manifest_is_empty(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_load_manifest_is_empty(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(os.path.isdir(agent.get_agent_dir()))
         agent._unpack()
@@ -462,7 +591,8 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
-    def test_load_manifest_is_malformed(self, mock_ensure):
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
+    def test_load_manifest_is_malformed(self, mock_loaded, mock_downloaded):
         agent = GuestAgent(path=self.agent_path)
         self.assertFalse(os.path.isdir(agent.get_agent_dir()))
         agent._unpack()
@@ -482,8 +612,9 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
     @patch("azurelinuxagent.ga.update.restutil.http_get")
-    def test_download(self, mock_http_get, mock_ensure):
+    def test_download(self, mock_http_get, mock_loaded, mock_downloaded):
         self.remove_agents()
         self.assertFalse(os.path.isdir(self.agent_path))
 
@@ -499,8 +630,9 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
     @patch("azurelinuxagent.ga.update.restutil.http_get")
-    def test_download_fail(self, mock_http_get, mock_ensure):
+    def test_download_fail(self, mock_http_get, mock_loaded, mock_downloaded):
         self.remove_agents()
         self.assertFalse(os.path.isdir(self.agent_path))
 
@@ -516,13 +648,15 @@ class TestGuestAgent(UpdateTestCase):
         return
 
     @patch("azurelinuxagent.ga.update.GuestAgent._ensure_downloaded")
+    @patch("azurelinuxagent.ga.update.GuestAgent._ensure_loaded")
     @patch("azurelinuxagent.ga.update.restutil.http_get")
-    def test_download_fallback(self, mock_http_get, mock_ensure):
+    def test_download_fallback(self, mock_http_get, mock_loaded, mock_downloaded):
         self.remove_agents()
         self.assertFalse(os.path.isdir(self.agent_path))
 
         mock_http_get.return_value = ResponseMock(
-            status=restutil.httpclient.SERVICE_UNAVAILABLE)
+            status=restutil.httpclient.SERVICE_UNAVAILABLE,
+            response="")
 
         ext_uri = 'ext_uri'
         host_uri = 'host_uri'
@@ -549,8 +683,12 @@ class TestGuestAgent(UpdateTestCase):
                           return_value=True):
             self.assertRaises(UpdateError, agent._download)
             self.assertEqual(mock_http_get.call_count, 4)
+
             self.assertEqual(mock_http_get.call_args_list[2][0][0], ext_uri)
+
             self.assertEqual(mock_http_get.call_args_list[3][0][0], art_uri)
+            a, k = mock_http_get.call_args_list[3]
+            self.assertEqual(False, k['use_proxy'])
 
             # ensure fallback works as expected
             with patch.object(HostPluginProtocol,
@@ -558,8 +696,16 @@ class TestGuestAgent(UpdateTestCase):
                               return_value=[art_uri, {}]):
                 self.assertRaises(UpdateError, agent._download)
                 self.assertEqual(mock_http_get.call_count, 6)
+
+                a, k = mock_http_get.call_args_list[3]
+                self.assertEqual(False, k['use_proxy'])
+
                 self.assertEqual(mock_http_get.call_args_list[4][0][0], ext_uri)
+                a, k = mock_http_get.call_args_list[4]
+
                 self.assertEqual(mock_http_get.call_args_list[5][0][0], art_uri)
+                a, k = mock_http_get.call_args_list[5]
+                self.assertEqual(False, k['use_proxy'])
 
     @patch("azurelinuxagent.ga.update.restutil.http_get")
     def test_ensure_downloaded(self, mock_http_get):
@@ -593,7 +739,7 @@ class TestGuestAgent(UpdateTestCase):
 
     @patch("azurelinuxagent.ga.update.GuestAgent._download")
     @patch("azurelinuxagent.ga.update.GuestAgent._unpack", side_effect=UpdateError)
-    def test_ensure_downloaded_unpack_fails(self, mock_download, mock_unpack):
+    def test_ensure_downloaded_unpack_fails(self, mock_unpack, mock_download):
         self.assertFalse(os.path.isdir(self.agent_path))
 
         pkg = ExtHandlerPackage(version=str(get_agent_version()))
@@ -608,7 +754,7 @@ class TestGuestAgent(UpdateTestCase):
     @patch("azurelinuxagent.ga.update.GuestAgent._download")
     @patch("azurelinuxagent.ga.update.GuestAgent._unpack")
     @patch("azurelinuxagent.ga.update.GuestAgent._load_manifest", side_effect=UpdateError)
-    def test_ensure_downloaded_load_manifest_fails(self, mock_download, mock_unpack, mock_manifest):
+    def test_ensure_downloaded_load_manifest_fails(self, mock_manifest, mock_unpack, mock_download):
         self.assertFalse(os.path.isdir(self.agent_path))
 
         pkg = ExtHandlerPackage(version=str(get_agent_version()))
@@ -623,10 +769,13 @@ class TestGuestAgent(UpdateTestCase):
     @patch("azurelinuxagent.ga.update.GuestAgent._download")
     @patch("azurelinuxagent.ga.update.GuestAgent._unpack")
     @patch("azurelinuxagent.ga.update.GuestAgent._load_manifest")
-    def test_ensure_download_skips_blacklisted(self, mock_download, mock_unpack, mock_manifest):
+    def test_ensure_download_skips_blacklisted(self, mock_manifest, mock_unpack, mock_download):
         agent = GuestAgent(path=self.agent_path)
+        self.assertEqual(0, mock_download.call_count)
+
         agent.clear_error()
         agent.mark_failure(is_fatal=True)
+        self.assertTrue(agent.is_blacklisted)
 
         pkg = ExtHandlerPackage(version=str(get_agent_version()))
         pkg.uris.append(ExtHandlerPackageUri())
@@ -637,7 +786,6 @@ class TestGuestAgent(UpdateTestCase):
         self.assertTrue(agent.is_blacklisted)
         self.assertEqual(0, mock_download.call_count)
         self.assertEqual(0, mock_unpack.call_count)
-        self.assertEqual(0, mock_manifest.call_count)
         return
 
 
@@ -680,6 +828,12 @@ class TestUpdate(UpdateTestCase):
         self.event_patch.stop()
         return
 
+    def _create_protocol(self, count=5, versions=None):
+        latest_version = self.prepare_agents(count=count)
+        if versions is None or len(versions) <= 0:
+            versions = [latest_version]
+        return ProtocolMock(versions=versions)
+
     def _test_upgrade_available(
             self,
             base_version=FlexibleVersion(AGENT_VERSION),
@@ -687,12 +841,9 @@ class TestUpdate(UpdateTestCase):
             versions=None,
             count=5):
 
-        latest_version = self.prepare_agents(count=count)
-        if versions is None or len(versions) <= 0:
-            versions = [latest_version]
-
         if protocol is None:
-            protocol = ProtocolMock(versions=versions)
+            protocol = self._create_protocol(count=count, versions=versions)
+
         self.update_handler.protocol_util = protocol
         conf.get_autoupdate_gafamily = Mock(return_value=protocol.family)
 
@@ -702,6 +853,16 @@ class TestUpdate(UpdateTestCase):
         self.assertTrue(self._test_upgrade_available())
         return
 
+    def test_upgrade_available_will_refresh_goal_state(self):
+        protocol = self._create_protocol()
+        protocol.emulate_stale_goal_state()
+        self.assertTrue(self._test_upgrade_available(protocol=protocol))
+        self.assertEqual(2, protocol.call_counts["get_vmagent_manifests"])
+        self.assertEqual(1, protocol.call_counts["get_vmagent_pkgs"])
+        self.assertEqual(1, protocol.call_counts["update_goal_state"])
+        self.assertTrue(protocol.goal_state_forced)
+        return
+
     def test_get_latest_agent_excluded(self):
         self.prepare_agent(AGENT_VERSION)
         self.assertFalse(self._test_upgrade_available(
@@ -777,7 +938,7 @@ class TestUpdate(UpdateTestCase):
             v = a.version
         return
 
-    def _test_ensure_no_orphans(self, invocations=3, interval=ORPHAN_WAIT_INTERVAL):
+    def _test_ensure_no_orphans(self, invocations=3, interval=ORPHAN_WAIT_INTERVAL, pid_count=0):
         with patch.object(self.update_handler, 'osutil') as mock_util:
             # Note:
             # - Python only allows mutations of objects to which a function has
@@ -792,15 +953,20 @@ class TestUpdate(UpdateTestCase):
 
             mock_util.check_pid_alive = Mock(side_effect=iterator)
 
+            pid_files = self.update_handler._get_pid_files()
+            self.assertEqual(pid_count, len(pid_files))
+
             with patch('os.getpid', return_value=42):
                 with patch('time.sleep', return_value=None) as mock_sleep:
                     self.update_handler._ensure_no_orphans(orphan_wait_interval=interval)
+                    for pid_file in pid_files:
+                        self.assertFalse(os.path.exists(pid_file))
                     return mock_util.check_pid_alive.call_count, mock_sleep.call_count
         return
 
     def test_ensure_no_orphans(self):
         fileutil.write_file(os.path.join(self.tmp_dir, "0_waagent.pid"), ustr(41))
-        calls, sleeps = self._test_ensure_no_orphans(invocations=3)
+        calls, sleeps = self._test_ensure_no_orphans(invocations=3, pid_count=1)
         self.assertEqual(3, calls)
         self.assertEqual(2, sleeps)
         return
@@ -823,7 +989,8 @@ class TestUpdate(UpdateTestCase):
         with patch('os.kill') as mock_kill:
             calls, sleeps = self._test_ensure_no_orphans(
                                         invocations=4,
-                                        interval=3*GOAL_STATE_INTERVAL)
+                                        interval=3*GOAL_STATE_INTERVAL,
+                                        pid_count=1)
             self.assertEqual(3, calls)
             self.assertEqual(2, sleeps)
             self.assertEqual(1, mock_kill.call_count)
@@ -888,6 +1055,21 @@ class TestUpdate(UpdateTestCase):
         self.assertEqual(kept_agents, self.update_handler.agents)
         return
 
+    @patch('azurelinuxagent.common.protocol.wire.WireClient.get_host_plugin')
+    def test_get_host_plugin_returns_host_for_wireserver(self, mock_get_host):
+        protocol = WireProtocol('12.34.56.78')
+        mock_get_host.return_value = "faux host"
+        host = self.update_handler._get_host_plugin(protocol=protocol)
+        mock_get_host.assert_called_once()
+        self.assertEqual("faux host", host)
+
+    @patch('azurelinuxagent.common.protocol.wire.WireClient.get_host_plugin')
+    def test_get_host_plugin_returns_none_otherwise(self, mock_get_host):
+        protocol = MetadataProtocol()
+        host = self.update_handler._get_host_plugin(protocol=protocol)
+        mock_get_host.assert_not_called()
+        self.assertEqual(None, host)
+
     def test_get_latest_agent(self):
         latest_version = self.prepare_agents()
 
@@ -919,17 +1101,19 @@ class TestUpdate(UpdateTestCase):
         return
 
     def test_get_pid_files(self):
-        previous_pid_file, pid_file, = self.update_handler._get_pid_files()
-        self.assertEqual(None, previous_pid_file)
-        self.assertEqual("0_waagent.pid", os.path.basename(pid_file))
+        pid_files = self.update_handler._get_pid_files()
+        self.assertEqual(0, len(pid_files))
         return
 
     def test_get_pid_files_returns_previous(self):
         for n in range(1250):
             fileutil.write_file(os.path.join(self.tmp_dir, str(n)+"_waagent.pid"), ustr(n+1))
-        previous_pid_file, pid_file, = self.update_handler._get_pid_files()
-        self.assertEqual("1249_waagent.pid", os.path.basename(previous_pid_file))
-        self.assertEqual("1250_waagent.pid", os.path.basename(pid_file))
+        pid_files = self.update_handler._get_pid_files()
+        self.assertEqual(1250, len(pid_files))
+
+        pid_dir, pid_name, pid_re = self.update_handler._get_pid_parts()
+        for p in pid_files:
+            self.assertTrue(pid_re.match(os.path.basename(p)))
         return
 
     def test_is_clean_start_returns_true_when_no_sentinal(self):
@@ -971,27 +1155,27 @@ class TestUpdate(UpdateTestCase):
             self.assertTrue(self.update_handler._is_orphaned)
         return
 
-    def test_load_agents(self):
+    def test_find_agents(self):
         self.prepare_agents()
 
         self.assertTrue(0 <= len(self.update_handler.agents))
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
         self.assertEqual(len(get_agents(self.tmp_dir)), len(self.update_handler.agents))
         return
 
-    def test_load_agents_does_reload(self):
+    def test_find_agents_does_reload(self):
         self.prepare_agents()
 
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
         agents = self.update_handler.agents
 
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
         self.assertNotEqual(agents, self.update_handler.agents)
         return
 
-    def test_load_agents_sorts(self):
+    def test_find_agents_sorts(self):
         self.prepare_agents()
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
 
         v = FlexibleVersion("100000")
         for a in self.update_handler.agents:
@@ -1001,7 +1185,7 @@ class TestUpdate(UpdateTestCase):
 
     def test_purge_agents(self):
         self.prepare_agents()
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
 
         # Ensure at least three agents initially exist
         self.assertTrue(2 < len(self.update_handler.agents))
@@ -1013,7 +1197,7 @@ class TestUpdate(UpdateTestCase):
         # Reload and assert only the kept agents remain on disk
         self.update_handler.agents = kept_agents
         self.update_handler._purge_agents()
-        self.update_handler._load_agents()
+        self.update_handler._find_agents()
         self.assertEqual(
             [agent.version for agent in kept_agents],
             [agent.version for agent in self.update_handler.agents])
@@ -1031,7 +1215,7 @@ class TestUpdate(UpdateTestCase):
             self.assertTrue(os.path.exists(agent_path + ".zip"))
         return
 
-    def _test_run_latest(self, mock_child=None, mock_time=None):
+    def _test_run_latest(self, mock_child=None, mock_time=None, child_args=None):
         if mock_child is None:
             mock_child = ChildMock()
         if mock_time is None:
@@ -1039,8 +1223,8 @@ class TestUpdate(UpdateTestCase):
 
         with patch('subprocess.Popen', return_value=mock_child) as mock_popen:
             with patch('time.time', side_effect=mock_time.time):
-                with patch('time.sleep', return_value=mock_time.sleep):
-                    self.update_handler.run_latest()
+                with patch('time.sleep', side_effect=mock_time.sleep):
+                    self.update_handler.run_latest(child_args=child_args)
                     self.assertEqual(1, mock_popen.call_count)
 
                     return mock_popen.call_args
@@ -1050,16 +1234,32 @@ class TestUpdate(UpdateTestCase):
 
         agent = self.update_handler.get_latest_agent()
         args, kwargs = self._test_run_latest()
+        args = args[0]
         cmds = textutil.safe_shlex_split(agent.get_agent_cmd())
         if cmds[0].lower() == "python":
             cmds[0] = get_python_cmd()
 
-        self.assertEqual(args[0], cmds)
+        self.assertEqual(args, cmds)
+        self.assertTrue(len(args) > 1)
+        self.assertTrue(args[0].startswith("python"))
+        self.assertEqual("-run-exthandlers", args[len(args)-1])
         self.assertEqual(True, 'cwd' in kwargs)
         self.assertEqual(agent.get_agent_dir(), kwargs['cwd'])
         self.assertEqual(False, '\x00' in cmds[0])
         return
 
+    def test_run_latest_passes_child_args(self):
+        self.prepare_agents()
+
+        agent = self.update_handler.get_latest_agent()
+        args, kwargs = self._test_run_latest(child_args="AnArgument")
+        args = args[0]
+
+        self.assertTrue(len(args) > 1)
+        self.assertTrue(args[0].startswith("python"))
+        self.assertEqual("AnArgument", args[len(args)-1])
+        return
+
     def test_run_latest_polls_and_waits_for_success(self):
         mock_child = ChildMock(return_value=None)
         mock_time = TimeMock(time_increment=CHILD_HEALTH_INTERVAL/3)
@@ -1082,7 +1282,22 @@ class TestUpdate(UpdateTestCase):
         self._test_run_latest(mock_child=mock_child, mock_time=mock_time)
         self.assertEqual(1, mock_child.poll.call_count)
         self.assertEqual(0, mock_child.wait.call_count)
-        self.assertEqual(2, mock_time.time_call_count)
+        return
+
+    def test_run_latest_polls_frequently_if_installed_is_latest(self):
+        mock_child = ChildMock(return_value=0)
+        mock_time = TimeMock(time_increment=CHILD_HEALTH_INTERVAL/2)
+        self._test_run_latest(mock_time=mock_time)
+        self.assertEqual(1, mock_time.sleep_interval)
+        return
+
+    def test_run_latest_polls_moderately_if_installed_not_latest(self):
+        self.prepare_agents()
+
+        mock_child = ChildMock(return_value=0)
+        mock_time = TimeMock(time_increment=CHILD_HEALTH_INTERVAL/2)
+        self._test_run_latest(mock_time=mock_time)
+        self.assertNotEqual(1, mock_time.sleep_interval)
         return
 
     def test_run_latest_defaults_to_current(self):
@@ -1131,13 +1346,13 @@ class TestUpdate(UpdateTestCase):
         with patch('azurelinuxagent.ga.update.UpdateHandler.get_latest_agent', return_value=latest_agent):
             self._test_run_latest(mock_child=ChildMock(return_value=1))
 
-        self.assertTrue(latest_agent.is_available)
+        self.assertTrue(latest_agent.is_blacklisted)
+        self.assertFalse(latest_agent.is_available)
         self.assertNotEqual(0.0, latest_agent.error.last_failure)
         self.assertEqual(1, latest_agent.error.failure_count)
         return
 
     def test_run_latest_exception_blacklists(self):
-        # logger.add_logger_appender(logger.AppenderType.STDOUT)
         self.prepare_agents()
 
         latest_agent = self.update_handler.get_latest_agent()
@@ -1154,6 +1369,24 @@ class TestUpdate(UpdateTestCase):
         self.assertEqual(1, latest_agent.error.failure_count)
         return
 
+    def test_run_latest_exception_does_not_blacklist_if_terminating(self):
+        self.prepare_agents()
+
+        latest_agent = self.update_handler.get_latest_agent()
+        self.assertTrue(latest_agent.is_available)
+        self.assertEqual(0.0, latest_agent.error.last_failure)
+        self.assertEqual(0, latest_agent.error.failure_count)
+
+        with patch('azurelinuxagent.ga.update.UpdateHandler.get_latest_agent', return_value=latest_agent):
+            self.update_handler.running = False
+            self._test_run_latest(mock_child=ChildMock(side_effect=Exception("Attempt blacklisting")))
+
+        self.assertTrue(latest_agent.is_available)
+        self.assertFalse(latest_agent.error.is_blacklisted)
+        self.assertEqual(0.0, latest_agent.error.last_failure)
+        self.assertEqual(0, latest_agent.error.failure_count)
+        return
+
     @patch('signal.signal')
     def test_run_latest_captures_signals(self, mock_signal):
         self._test_run_latest()
@@ -1183,8 +1416,6 @@ class TestUpdate(UpdateTestCase):
                 self.update_handler.running = False
             return
 
-        calls = calls * invocations
-
         fileutil.write_file(conf.get_agent_pid_file_path(), ustr(42))
 
         with patch('azurelinuxagent.ga.exthandlers.get_exthandlers_handler') as mock_handler:
@@ -1212,7 +1443,7 @@ class TestUpdate(UpdateTestCase):
         return
 
     def test_run_keeps_running(self):
-        self._test_run(invocations=15)
+        self._test_run(invocations=15, calls=[call.run()]*15)
         return
 
     def test_run_stops_if_update_available(self):
@@ -1277,12 +1508,14 @@ class TestUpdate(UpdateTestCase):
     def test_shutdown(self):
         self.update_handler._set_sentinal()
         self.update_handler._shutdown()
+        self.assertFalse(self.update_handler.running)
         self.assertFalse(os.path.isfile(self.update_handler._sentinal_file_path()))
         return
 
     def test_shutdown_ignores_missing_sentinal_file(self):
         self.assertFalse(os.path.isfile(self.update_handler._sentinal_file_path()))
         self.update_handler._shutdown()
+        self.assertFalse(self.update_handler.running)
         self.assertFalse(os.path.isfile(self.update_handler._sentinal_file_path()))
         return
 
@@ -1300,8 +1533,9 @@ class TestUpdate(UpdateTestCase):
         for n in range(1112):
             fileutil.write_file(os.path.join(self.tmp_dir, str(n)+"_waagent.pid"), ustr(n+1))
         with patch('os.getpid', return_value=1112):
-            previous_pid_file, pid_file = self.update_handler._write_pid_file()
-            self.assertEqual("1111_waagent.pid", os.path.basename(previous_pid_file))
+            pid_files, pid_file = self.update_handler._write_pid_file()
+            self.assertEqual(1112, len(pid_files))
+            self.assertEqual("1111_waagent.pid", os.path.basename(pid_files[-1]))
             self.assertEqual("1112_waagent.pid", os.path.basename(pid_file))
             self.assertEqual(fileutil.read_file(pid_file), ustr(1112))
         return
@@ -1309,8 +1543,8 @@ class TestUpdate(UpdateTestCase):
     def test_write_pid_file_ignores_exceptions(self):
         with patch('azurelinuxagent.common.utils.fileutil.write_file', side_effect=Exception):
             with patch('os.getpid', return_value=42):
-                previous_pid_file, pid_file = self.update_handler._write_pid_file()
-                self.assertEqual(None, previous_pid_file)
+                pid_files, pid_file = self.update_handler._write_pid_file()
+                self.assertEqual(0, len(pid_files))
                 self.assertEqual(None, pid_file)
         return
 
@@ -1328,12 +1562,22 @@ class ProtocolMock(object):
     def __init__(self, family="TestAgent", etag=42, versions=None, client=None):
         self.family = family
         self.client = client
+        self.call_counts = {
+            "get_vmagent_manifests" : 0,
+            "get_vmagent_pkgs" : 0,
+            "update_goal_state" : 0
+        }
+        self.goal_state_is_stale = False
+        self.goal_state_forced = False
         self.etag = etag
         self.versions = versions if versions is not None else []
         self.create_manifests()
         self.create_packages()
         return
 
+    def emulate_stale_goal_state(self):
+        self.goal_state_is_stale = True
+
     def create_manifests(self):
         self.agent_manifests = VMAgentManifestList()
         if len(self.versions) <= 0:
@@ -1364,16 +1608,29 @@ class ProtocolMock(object):
         return self
 
     def get_vmagent_manifests(self):
+        self.call_counts["get_vmagent_manifests"] += 1
+        if self.goal_state_is_stale:
+            self.goal_state_is_stale = False
+            raise ResourceGoneError()
         return self.agent_manifests, self.etag
 
     def get_vmagent_pkgs(self, manifest):
+        self.call_counts["get_vmagent_pkgs"] += 1
+        if self.goal_state_is_stale:
+            self.goal_state_is_stale = False
+            raise ResourceGoneError()
         return self.agent_packages
 
+    def update_goal_state(self, forced=False, max_retry=3):
+        self.call_counts["update_goal_state"] += 1
+        self.goal_state_forced = self.goal_state_forced or forced
+        return
 
 class ResponseMock(Mock):
-    def __init__(self, status=restutil.httpclient.OK, response=None):
+    def __init__(self, status=restutil.httpclient.OK, response=None, reason=None):
         Mock.__init__(self)
         self.status = status
+        self.reason = reason
         self.response = response
         return
 
@@ -1388,7 +1645,11 @@ class TimeMock(Mock):
         self.time_call_count = 0
         self.time_increment = time_increment
 
-        self.sleep = Mock(return_value=0)
+        self.sleep_interval = None
+        return
+
+    def sleep(self, n):
+        self.sleep_interval = n
         return
 
     def time(self):
@@ -1397,6 +1658,5 @@ class TimeMock(Mock):
         self.next_time += self.time_increment
         return current_time
 
-
 if __name__ == '__main__':
     unittest.main()
